import configparser
import json
import os
import re
import time
from typing import Literal, cast, overload

import requests

from splatnet3_scraper import __version__
from splatnet3_scraper.base.exceptions import (
    NintendoException,
    SplatnetException,
)
from splatnet3_scraper.base.graph_ql_queries import queries
from splatnet3_scraper.base.tokens.nso import NSO
from splatnet3_scraper.constants import (
    ENV_VAR_NAMES,
    GRAPH_QL_REFERENCE_URL,
    TOKEN_EXPIRATIONS,
    TOKENS,
)
from splatnet3_scraper.utils import retry

text_config_re = re.compile(r"\s*=*\s*")


class Token:
    """Represents at token. Contains the token itself, the type of token, and
    the time it was created. Can be used to check if the token is expired or
    display the time left before it expires.
    """

    def __init__(self, token: str, token_type: str, timestamp: float) -> None:
        self.token = token
        self.token_type = token_type
        self.timestamp = timestamp
        self.expiration = TOKEN_EXPIRATIONS.get(token_type, 1e10) + timestamp

    @property
    def is_valid(self) -> bool:
        return (self.token is not None) and (self.token != "")

    @property
    def is_expired(self) -> bool:
        return self.time_left <= 0

    @property
    def time_left(self) -> float:
        return self.expiration - time.time()

    @property
    def time_left_str(self) -> str:
        """Returns a string representation of the time left before the token
        expires. If the token is expired, "Expired" will be returned.

        Returns:
            str: A string representation of the time left before the token
        """
        time_left = self.time_left
        if time_left <= 0:
            return "Expired"
        mins, secs = divmod(time_left, 60)
        hours, mins = divmod(mins, 60)

        out = ""
        if hours > 1e5:
            return "basically forever"
        if hours > 0:
            out += f"{hours:.0f}h "
        if mins > 0:
            out += f"{mins:.0f}m "
        if secs > 0:
            out += f"{secs:.1f}s"
        return out.strip()

    def __repr__(self) -> str:
        out = "Token("
        spaces = " " * len(out)
        out += (
            f"token={self.token[:5]}...,\n"
            + spaces
            + f"type={self.token_type},\n"
            + spaces
            + "expires in "
            + self.time_left_str
            + "\n)"
        )
        return out


class TokenManager:
    """Manages tokens. Can be used to add tokens, generate tokens from the NSO
    class, check if tokens are expired, load tokens from a config file or
    environment variables, save tokens to a config file, and display the time
    left before tokens expire.
    """

    def __init__(self, nso: NSO | None = None) -> None:
        nso = nso if nso is not None else NSO.new_instance()
        self.nso = nso
        self._tokens: dict[str, Token] = {}
        self._data: dict[str, str] = {}

    def flag_origin(self, origin: str, data: str | None = None) -> None:
        """Flags the origin of the token manager. This is used to determine
        whether the token manager was loaded from a config file or environment
        variables.

        Args:
            origin (str): The origin of the token manager.
            data (str | None): Additional data about the origin. For example,
                if the token manager was loaded from a config file, this would
                be the path to the config file. On the other hand, if the token
                manager was loaded from environment variables, this would be
                None.
        """
        self._origin = {"origin": origin, "data": data}

    def add_token(
        self,
        token: str | Token,
        token_type: str | None = None,
        timestamp: float | None = None,
    ) -> None:
        """Adds a token to the manager. If the token is a string, the token
        type must be provided. If the token is a Token object, the object will
        be added to the manager.

        Args:
            token (str | Token): The token to add.
            token_type (str | None): The type of token. If token is an instance
                of Token, this will be ignored. If token is a string, this must
                be provided.
            timestamp (float | None): The time the token was created. If not
                provided, the current time will be used.

        Raises:
            ValueError: If token is a string and token_type is not provided.
        """
        if isinstance(token, Token):
            self._tokens[token.token_type] = token
            if token.token_type == TOKENS.GTOKEN:
                self.nso._gtoken = token.token
            return
        if token_type is None:
            raise ValueError("token_type must be provided if token is a str.")
        if timestamp is None:
            timestamp = time.time()
        token_obj = Token(token, token_type, timestamp)
        if token_obj.token_type == TOKENS.GTOKEN:
            self.nso._gtoken = token_obj.token
        self._tokens[token_type] = token_obj

    @overload
    def get(self, token_type: str, full_token: Literal[False] = ...) -> str:
        ...

    @overload
    def get(self, token_type: str, full_token: Literal[True]) -> Token:
        ...

    @overload
    def get(self, token_type: str, full_token: bool) -> str | Token:
        ...

    def get(self, token_type: str, full_token: bool = False) -> str | Token:
        """Gets a token from the manager. If full_token is True, the Token
        object will be returned. Otherwise, the token string will be returned.

        Args:
            token_type (str): The type of token to get.
            full_token (bool): Whether to return the full Token object or just
                the token string.

        Raises:
            ValueError: If the token type is not found.

        Returns:
            str | Token: The token or Token object.
        """
        token_obj = self._tokens.get(token_type, None)
        if token_obj is None:
            raise ValueError(f"Token of type {token_type} not found.")
        if full_token:
            return token_obj
        return token_obj.token

    @property
    def data(self) -> dict[str, str]:
        """Returns the data stored in the manager.

        Returns:
            dict[str, str]: The data stored in the manager.
        """
        return self._data

    def add_session_token(self, token: str) -> None:
        """Adds a session token to the manager.

        Args:
            token (str): The session token to add.
        """
        self.add_token(token, TOKENS.SESSION_TOKEN)
        self.nso._session_token = token

    def generate_gtoken(self) -> None:
        """Generates a gtoken from the NSO class and adds it to the manager.
        Requires a session token to already be set.

        Raises:
            ValueError: If the session token has not been set.
            NintendoException: If the user info could not be retrieved.
        """
        if TOKENS.SESSION_TOKEN not in self._tokens:
            raise ValueError(
                "Session token must be set before generating a gtoken."
            )
        gtoken = self.nso.get_gtoken(self.nso.session_token)
        self.add_token(gtoken, TOKENS.GTOKEN)
        try:
            user_info = cast(dict[str, str], self.nso._user_info)
            country = user_info["country"]
            language = user_info["language"]
            self._data["country"] = country
            self._data["language"] = language
        except (KeyError, TypeError):
            raise NintendoException(
                "Unable to get user info. Gtoken may be invalid."
            )

    @retry(times=1, exceptions=SplatnetException)
    def generate_bullet_token(self) -> None:
        """Generates a bullet token from the NSO class and adds it to the
        manager. If a gtoken has not been generated, one will be generated
        before generating the bullet token. Requires a session token to already
        be set.

        Raises:
            ValueError: If the session token has not been set.
            SplatnetException: If the bullet token was unable to be generated.
        """
        if TOKENS.SESSION_TOKEN not in self._tokens:
            raise ValueError(
                "Session token must be set before generating a bullet token."
            )
        if (TOKENS.GTOKEN not in self._tokens) or (self.nso._user_info is None):
            self.generate_gtoken()
        bullet_token = self.nso.get_bullet_token(
            cast(str, self.nso._gtoken), cast(dict, self.nso._user_info)
        )
        self.add_token(bullet_token, TOKENS.BULLET_TOKEN)
        bullet = self.get(TOKENS.BULLET_TOKEN, full_token=True)
        if (bullet is not None) and not bullet.is_valid:
            raise SplatnetException(
                "Bullet token was unable to be generated. This is likely due "
                "to SplatNet 3 being down. Please try again later."
            )

    def generate_all_tokens(self) -> None:
        """Generates all tokens from the NSO class and adds them to the
        manager. Requires a session token to already be set.
        """
        self.generate_gtoken()
        self.generate_bullet_token()

    @classmethod
    def from_session_token(cls, session_token: str) -> "TokenManager":
        """Creates a token manager from a session token.

        Args:
            session_token (str): The session token to use.

        Returns:
            TokenManager: The token manager with the session token added.
        """
        manager = cls()
        manager.add_session_token(session_token)
        manager.flag_origin("session_token")
        return manager

    @classmethod
    def load(cls) -> "TokenManager":
        """Loads tokens from a config file or environment variables.

        Checks for appropriate tokens in the following order:
            1. .splatnet3_scraper file
            2. Environment variables
            3. tokens.ini file

        Raises:
            ValueError: If no tokens are found.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        if os.path.exists(".splatnet3_scraper"):
            return cls.from_config_file(".splatnet3_scraper")
        elif any([os.environ.get(var) for var in ENV_VAR_NAMES.values()]):
            return cls.from_env()
        elif os.path.exists("tokens.ini"):
            return cls.from_config_file("tokens.ini")
        else:
            raise ValueError(
                "No tokens found. Please create a .splatnet3_scraper file, set "
                "environment variables, or create a tokens.ini file."
            )

    @classmethod
    def from_config_file(cls, path: str) -> "TokenManager":
        """Loads tokens from a config file.

        Args:
            path (str): The path to the config file.

        Raises:
            ValueError: If the config file does not have a 'tokens' section.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        config = configparser.ConfigParser()
        config.read(path)
        nso = NSO.new_instance()
        tokenmanager = cls(nso)
        tokenmanager.flag_origin("config_file", path)

        if not config.has_section("tokens"):
            raise ValueError("Config file does not have a 'tokens' section.")
        for option in config.options("tokens"):
            token = config.get("tokens", option)
            if option == TOKENS.SESSION_TOKEN:
                nso._session_token = token
            elif option == TOKENS.GTOKEN:
                nso._gtoken = token
            tokenmanager.add_token(token, option)

        if not config.has_section("data"):
            tokenmanager.generate_all_tokens()
            return tokenmanager
        for option in config.options("data"):
            tokenmanager._data[option] = config.get("data", option)
        tokenmanager.test_tokens()
        return tokenmanager

    @classmethod
    def from_text_file(cls, path: str) -> "TokenManager":
        """Loads tokens from a text file. Not recommended, but here for
        compatability with s3s config files.

        Args:
            path (str): The path to the text file.

        Raises:
            ValueError: If the session token is not found in the text file.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        token_manager = cls()
        with open(path, "r") as f:
            data = json.load(f)

        if "session_token" not in data:
            raise ValueError("Session token not found in text file.")
        token_manager.add_session_token(data["session_token"])
        token_manager.flag_origin("text_file", path)
        if "acc_loc" in data:
            language, country = data["acc_loc"].split("|")
            token_manager._data["language"] = language
            token_manager._data["country"] = country

        if "gtoken" in data:
            token_manager.add_token(data["gtoken"], TOKENS.GTOKEN)
        if "bullettoken" in data:
            token_manager.add_token(data["bullettoken"], TOKENS.BULLET_TOKEN)
        token_manager.test_tokens()
        return token_manager

    @classmethod
    def from_env(cls) -> "TokenManager":
        """Loads tokens from environment variables.

        Raises:
            ValueError: If the session token environment variable is not set.

        Returns:
            TokenManager: The token manager with the tokens loaded.
        """
        nso = NSO.new_instance()
        tokenmanager = cls(nso)
        for token in ENV_VAR_NAMES:
            token_env = os.environ.get(ENV_VAR_NAMES[token])
            if token == TOKENS.SESSION_TOKEN:
                if token_env is None:
                    raise ValueError(
                        "Session token environment variable not set."
                    )
                tokenmanager.nso._session_token = token_env
            elif token_env is None:
                continue
            elif token == TOKENS.GTOKEN:
                tokenmanager.nso._gtoken = token_env
            tokenmanager.add_token(token_env, token)
        tokenmanager.flag_origin("env")
        tokenmanager.test_tokens()
        return tokenmanager

    def save(self, path: str | None = None) -> None:
        """Saves the tokens to a config file.

        Args:
            path (str): The path to the config file.
        """
        config = configparser.ConfigParser()
        out_tokens = {}
        for token_name, token in self._tokens.items():
            out_tokens[token_name] = token.token
        config["tokens"] = out_tokens
        config["data"] = self._data
        config["metadata"] = {
            "version": __version__,
            "class": self.__class__.__name__,
        }
        if path is None:
            path = ".splatnet3_scraper"
        with open(path, "w") as configfile:
            config.write(configfile)

    def token_is_valid(self, token_type: str) -> bool:
        """Checks if a token is valid.

        Args:
            token_type (str): The type of token to check.

        Returns:
            bool: True if the token is valid, False otherwise.
        """
        try:
            token = self.get(token_type, full_token=True)
        except ValueError:
            return False
        return token.is_valid

    def test_tokens(self, user_agent: str | None = None) -> None:
        """Tests the tokens by making a request to the GraphQL endpoint and
        regenerate tokens if they are invalid.

        Args:
            user_agent (str): The user agent to use for the request.

        Raises:
            ValueError: If the session token is not set.
        """
        if self.get(TOKENS.SESSION_TOKEN) is None:
            raise ValueError("Session Token is not set.")

        if self.token_is_valid(TOKENS.GTOKEN) is False:
            self.generate_gtoken()

        if self.token_is_valid(TOKENS.BULLET_TOKEN) is False:
            self.generate_bullet_token()

        header = queries.query_header(
            self.get(TOKENS.BULLET_TOKEN), self._data["language"], user_agent
        )

        response = requests.post(
            GRAPH_QL_REFERENCE_URL,
            data=queries.query_body("HomeQuery"),
            headers=header,
            cookies={"_gtoken": cast(str, self.get(TOKENS.GTOKEN))},
        )
        if response.status_code != 200:
            self.generate_all_tokens()



import json
from typing import ParamSpec, TypeVar

import requests

from splatnet3_scraper.base.exceptions import SplatnetException
from splatnet3_scraper.base.graph_ql_queries import queries
from splatnet3_scraper.base.tokens import NSO, TokenManager
from splatnet3_scraper.constants import TOKENS
from splatnet3_scraper.scraper.config import Config
from splatnet3_scraper.scraper.query_map import QueryMap
from splatnet3_scraper.scraper.responses import QueryResponse
from splatnet3_scraper.utils import retry

T = TypeVar("T")
P = ParamSpec("P")


class SplatNet3_Scraper:
    """This class implements all of the classes and functions defined in the
    splatnet3_scraper.base module and can be seen as an example of how to put
    them together. While the class has a method to create the session token, a
    full implementation with a CLI is outside the scope of this project.
    """

    def __init__(self, config: Config) -> None:
        """Initializes the class, it is not meant to be instantiated directly,
        but rather through one of the available factory methods. Still, it can
        be instantiated directly if the user wants to use a custom Config class.

        Args:
            config (Config): The configuration to use.
        """
        self.config = config

    @staticmethod
    def from_config_file(config_file: str | None = None) -> "SplatNet3_Scraper":
        """Creates a new instance of the class using a configuration file. If
        no configuration file is provided, the default configuration file,
        .splatnet3_scraper, will be used.

        Args:
            config_file (str | None): The path to the configuration file. If
                None, the default configuration file will be used. Defaults to
                None.

        Returns:
            SplatNet3_Scraper: A new instance of the class.
        """
        config = Config(config_file)
        return SplatNet3_Scraper(config)

    @staticmethod
    def generate_session_token_url(
        user_agent: str | None = None,
    ) -> tuple[str, bytes, bytes]:
        """Generates the URL to use to get the session token.

        Args:
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.

        Returns:
            tuple:
                str: The URL to use to get the session token.
                bytes: The state ID.
                bytes: The challenge solution.
        """
        nso = NSO.new_instance()
        url = nso.generate_login_url(user_agent)
        return url, nso.state, nso.verifier

    @staticmethod
    def from_session_token(session_token: str) -> "SplatNet3_Scraper":
        """Creates a new instance of the class using a session token.

        Args:
            session_token (str): The session token.

        Returns:
            SplatNet3_Scraper: A new instance of the class.
        """
        token_manager = TokenManager.from_session_token(session_token)
        token_manager.generate_all_tokens()
        config = Config(token_manager=token_manager)
        return SplatNet3_Scraper(config)

    @staticmethod
    def from_env() -> "SplatNet3_Scraper":
        """Creates a new instance of the class using the environment
        variables.

        Returns:
            SplatNet3_Scraper: A new instance of the class.
        """
        config = Config.from_env()
        return SplatNet3_Scraper(config)

    @staticmethod
    def from_s3s_config(path: str) -> "SplatNet3_Scraper":
        """Creates a new instance of the class using the s3s config file.

        Args:
            path (str): The path to the s3s config file.

        Returns:
            SplatNet3_Scraper: A new instance of the class.
        """
        config = Config.from_s3s_config(path)
        return SplatNet3_Scraper(config)

    def __query(
        self, query_name: str, variables: dict = {}
    ) -> requests.Response:
        """Internal method to query Splatnet 3, it does all of the heavy lifting
        and is used by the other methods to get the data.

        Args:
            query_name (str): The name of the query to use.
            variables (dict): The variables to use in the query. Defaults to {}.

        Returns:
            requests.Response: The response from the query.
        """
        return queries.query(
            query_name,
            self.config.get_token(TOKENS.BULLET_TOKEN),
            self.config.get_token(TOKENS.GTOKEN),
            self.config.get_data("language"),
            self.config.get("user_agent"),
            variables=variables,
        )

    def __get_query(self, query_name: str) -> dict:
        """Internal method to get the data from a query. If the query fails, it
        will try to generate all of the tokens and only once more. Not
        recommended to wrap this method in the retry decorator generator.

        Args:
            query_name (str): The name of the query to use.

        Returns:
            dict: The data from the query.
        """
        response = self.__query(query_name)
        if response.status_code != 200:
            self.config.token_manager.generate_all_tokens()
            response = self.__query(query_name)
        return response.json()["data"]

    def __get_detailed_query(self, game_id: str, versus: bool = True) -> dict:
        """Internal method to get the data from a detailed query. If the query
        fails, it will try to generate all of the tokens and only once more.
        Not recommended to wrap this method in the retry decorator generator.

        Args:
            game_id (str): The ID of the game to get the details from.
            versus (bool): Whether the game is a versus game or not. Defaults
                to True.

        Returns:
            dict: The data from the query.
        """
        name = QueryMap.VS_DETAIL if versus else QueryMap.SALMON_DETAIL
        variable_name = "vsResultId" if versus else "coopHistoryDetailId"
        response = self.__query(name, variables={variable_name: game_id})
        if response.status_code != 200:
            self.config.token_manager.generate_all_tokens()
            response = self.__query(name, variables={variable_name: game_id})
        return response.json()["data"]

    def get_mode_summary(
        self,
        mode: str,
        detailed: bool = False,
        detailed_limit: int | None = None,
    ) -> QueryResponse:
        """Gets the summary for a specific mode.

        Args:
            mode (str): The mode to get the summary for.
            detailed (bool): Whether to get the detailed summary or not.
                Defaults to False.
            detailed_limit (int | None): The maximum number of detailed
                matches to get. If None, all of the matches will be returned.

        Raises:
            ValueError: If the mode is invalid.

        Returns:
            QueryResponse: The summary for the mode.
        """
        try:
            query_name = QueryMap.get(mode)
        except AttributeError:
            raise ValueError(f"Invalid mode: {mode}")

        if not detailed:
            data = self.__get_query(query_name)
            return QueryResponse(data=data)
        summary, detailed_data = self.__vs_with_details(
            query_name, detailed_limit
        )
        return QueryResponse(data=summary, additional_data=detailed_data)

    def get_vs_detail(self, game_id: str) -> dict:
        """Gets the details of a versus game.

        Args:
            game_id (str): The ID of the game to get the details from.

        Returns:
            dict: The details of the game.
        """
        return self.__get_detailed_query(game_id)

    def get_salmon_detail(self, game_id: str) -> dict:
        """Gets the details of a salmon run game.

        Args:
            game_id (str): The ID of the game to get the details from.

        Returns:
            dict: The details of the game.
        """
        return self.__get_detailed_query(game_id, False)

    def __vs_with_details(
        self, query_name: str, limit: int | None = None
    ) -> tuple[dict, list[dict]]:
        """Internal method to get the data from a query and add the details to
        the data. If the query fails, it will try to generate all of the tokens
        and only once more. Not recommended to wrap this method in the retry
        decorator generator.

        Args:
            query_name (str): The name of the query to use.
            limit (int | None): The maximum number of detailed matches to get.
                If None, all of the matches will be returned. Defaults to None.

        Returns:
            tuple:
                dict: The data from the query.
                list[dict]: The details of the games, in order.
        """
        data = self.__get_query(query_name)
        if limit is None:
            limit = -1
        # Top level key depends on the game mode but there is only one top level
        # key so we can just get the first one.
        key = list(data.keys())[0]
        base = data[key]["historyGroups"]["nodes"]
        out: list[dict] = []
        idx = 0
        for group in base:
            for game in group["historyDetails"]["nodes"]:
                if idx == limit:
                    return data, out
                idx += 1
                game_id = game["id"]
                detailed_game = self.get_vs_detail(game_id)
                out.append(detailed_game)
        return data, out

    @retry(times=1, exceptions=ConnectionError)
    def query(self, query_name: str, variables: dict = {}) -> dict:
        """Queries Splatnet 3 and returns the data.

        Args:
            query_name (str): The name of the query to use.
            variables (dict): The variables to use in the query. Defaults to {}.

        Raises:
            SplatnetException: If the query was successful but returned at
                least one error.

        Returns:
            dict: The data from the query.
        """
        response = self.__query(query_name, variables)
        if response.status_code != 200:
            self.config.token_manager.generate_all_tokens()
            response = self.__query(query_name, variables)

        if "errors" in response.json():
            errors = response.json()["errors"]
            error_message = (
                "Query was successful but returned at least one error."
            )

            error_message += " Errors: " + json.dumps(errors, indent=4)
            raise SplatnetException(error_message)

        return QueryResponse(data=response.json()["data"])



import base64
import hashlib
import json
import os
import re
from typing import cast

import requests

from splatnet3_scraper import __version__
from splatnet3_scraper.base.exceptions import (
    FTokenException,
    NintendoException,
    SplatnetException,
)
from splatnet3_scraper.constants import (
    APP_VERSION_FALLBACK,
    DEFAULT_USER_AGENT,
    IMINK_URL,
    IOS_APP_URL,
    SPLATNET_URL,
    WEB_VIEW_VERSION_FALLBACK,
)
from splatnet3_scraper.logs import logger
from splatnet3_scraper.utils import get_splatnet_web_version, retry

version_re = re.compile(
    r"(?<=whats\-new\_\_latest\_\_version\"\>Version)\s+\d+\.\d+\.\d+"
)


class NSO:
    """The NSO class contains all the logic to proceed through the login flow.
    This class also holds various properties that are used to make requests to
    the Nintendo Switch Online API. Login flow is roughly as follows:
        1.  Initialize a requests session and store it.
        2.  Generate a random state and S256 code challenge that will be used
                to obtain the "session_token". Store them for later use.
        3.  Generate a login URL using the state and code challenge that the
                user will open in their browser. The user will then copy a link
                and feed it back to the program.
        4.  Parse the URI to obtain the session token code, then use it
                alongside the code challenge to obtain the "session_token".
                Store it for later use. The session token is valid for 2 years
                and can be revoked by the user.
        5.  Use the session token to obtain a user access response. Store it for
                later use. The user access response contains two tokens: an "id"
                token and a "user access" token.
        6.  Use the user access token to obtain the user information. This is
                required to obtain the "f" token. Store it for later use.
        7.  Use the user information to obtain an "f" token using the first
                hashing method. This will also return a request ID and a
                timestamp. These do not need to be stored.
        8.  Use the "f" token, user information, and the id token to obtain a
                splatoon token. This token will contain a new "id" token.
        9.  Use the new "id" token to obtain the g token using the second
                hashing method. Store it for later use. The g token is valid for
                6 hours and 30 minutes.
        10. Use the g token and user information to obtain a bullet token. This
                token is valid for 2 hours.

    Once the login flow is complete, the NSO class contains all the necessary
    values to regenerate the tokens. To minimize the number of objects to track,
    the NSO class only stores the tokens in memory. The TokenManager class
    handles the persistence of the tokens to disk.
    """

    def __init__(self, session: requests.Session) -> None:
        self.session = session
        self._state: bytes | None = None
        self._verifier: bytes | None = None
        self._version: str | None = None
        self._web_view_version: str | None = None
        self._session_token: str | None = None
        self._user_access_token: str | None = None
        self._id_token: str | None = None
        self._gtoken: str | None = None
        self._user_info: dict[str, str] | None = None

    @staticmethod
    def new_instance() -> "NSO":
        """Creates a new instance of the NSO class.

        The NSO class contains all the logic for obtaining an auth token from
        Nintendo. It also contains the session object that is used to make
        requests to the Nintendo Switch Online API, as well as the auth state
        and S256 code challenge used to obtain the auth token.

        Returns:
            NSO: A new instance of the NSO class.
        """
        session = requests.Session()
        return NSO(session=session)

    @property
    def version(self) -> str:
        """Returns the current version of the NSO app. Necessary to get the
        session token.

        Returns:
            str: The current version of the NSO app.
        """
        if self._version is None:
            self._version = self.get_version()
        return self._version

    @retry(times=2, exceptions=ValueError)
    def get_version(self) -> str:
        """Gets the current version of the NSO app. Necessary to access the API.
        Retries twice if the version cannot be obtained, in case the ios app
        store site is down or slow.

        Returns:
            str: The current version of the NSO app.
        """
        response = self.session.get(IOS_APP_URL)
        version = version_re.search(response.text)
        if version is None:
            logger.log("Failed to get version from app store, using fallback")
            return APP_VERSION_FALLBACK
        return version.group(0).strip()

    @property
    def state(self) -> bytes:
        """Returns a base64url encoded random 36 byte string for the auth state.

        Returns:
            bytes: The auth state.
        """
        if self._state is None:
            self._state = self.generate_new_state()
        return self._state

    def generate_new_state(self) -> bytes:
        """Generates a new state.

        Returns:
            bytes: The auth state.
        """
        return base64.urlsafe_b64encode(os.urandom(36))

    @property
    def verifier(self) -> bytes:
        """Returns a base64url encoded random 32 byte string for the code
        verifier. This is used to generate the S256 code challenge. To align
        with node.js's crypto module, padding is removed.

        Returns:
            bytes: The code verifier, without padding.
        """
        if self._verifier is None:
            self._verifier = self.generate_new_verifier()
        return self._verifier

    def generate_new_verifier(self) -> bytes:
        """Generates a new verifier without padding.

        Returns:
            bytes: The code verifier, without padding.
        """
        verifier_ = base64.urlsafe_b64encode(os.urandom(32))
        return verifier_.replace(b"=", b"")

    @property
    def session_token(self) -> str:
        if self._session_token is None:
            raise ValueError("Session token is not set.")
        return self._session_token

    def header(self, user_agent: str) -> dict[str, str]:
        """Returns the headers for the session.

        Args:
            user_agent (str): The user agent to use.

        Returns:
            dict[str, str]: The headers.
        """
        return {
            "Host": "accounts.nintendo.com",
            "Connection": "keep-alive",
            "Cache-Control": "max-age=0",
            "Upgrade-Insecure-Requests": "1",
            "User-Agent": user_agent,
            "Accept": (
                "text/html,"
                + "application/xhtml+xml,"
                + "application/xml;q=0.9,"
                + "image/webp,"
                + "image/apng,"
                + "*/*;q=0.8n"
            ),
            "DNT": "1",
            "Accept-Encoding": "gzip,deflate,br",
        }

    def generate_login_url(self, user_agent: str | None = None) -> str:
        """Log in to a Nintendo account and returns a session token.

        Args:
            user_agent (str): The user agent to use. Defaults to the default
                user agent found in constants.py.

        Returns:
            str: The session token.
        """
        # https://dev.to/mathewthe2/intro-to-nintendo-switch-rest-api-2cm7
        hash_ = hashlib.sha256()
        hash_.update(self.verifier)
        challenge = base64.urlsafe_b64encode(hash_.digest()).replace(b"=", b"")

        header = self.header(
            user_agent=user_agent
            if user_agent is not None
            else DEFAULT_USER_AGENT
        )
        params = {
            "state": self.state,
            "redirect_uri": "npf71b963c1b7b6d119://auth",
            "client_id": "71b963c1b7b6d119",
            "scope": ("openid user user.birthday user.mii user.screenName"),
            "response_type": "session_token_code",
            "session_token_code_challenge": challenge,
            "session_token_code_challenge_method": "S256",
            "theme": "login_form",
        }
        login_url = "https://accounts.nintendo.com/connect/1.0.0/authorize"
        response = self.session.get(
            login_url, headers=header, params=params  # type: ignore
        )
        return response.url

    def parse_npf_uri(self, uri: str) -> str:
        """Parses the uri returned by the Nintendo login page and extracts the
        session token code. This is used to pass the challenge and verify that
        the user is who they say they are.

        Args:
            uri (str): The uri returned by the Nintendo login page.

        Returns:
            str: The session token code.
        """
        return uri.split("&")[1][len("session_token_code=") :]

    def get_session_token(self, session_token_code: str) -> str:
        """Gets the session token from the session token code.

        Args:
            session_token_code (str): The session token code.

        Returns:
            str: The session token.
        """
        header = {
            "User-Agent": f"OnlineLounge/{self.version} NASDKAPI Android",
            "Accept-Language": "en-US",
            "Accept": "application/json",
            "Content-Type": "application/x-www-form-urlencoded",
            "Content-Length": "540",
            "Host": "accounts.nintendo.com",
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
        }
        params = {
            "client_id": "71b963c1b7b6d119",
            "session_token_code": session_token_code,
            "session_token_code_verifier": self.verifier,
        }
        uri = "https://accounts.nintendo.com/connect/1.0.0/api/session_token"
        response = self.session.post(uri, headers=header, data=params)
        session_token = response.json()["session_token"]
        self._session_token = session_token
        return session_token

    def get_user_access_token(self, session_token: str) -> requests.Response:
        """Gets the user access token from the session token.

        Args:
            session_token (str): The session token.

        Returns:
            requests.Response: The response.
        """
        header = {
            "Host": "accounts.nintendo.com",
            "Accept-Encoding": "gzip",
            "Content-Type": "application/json",
            "Content-Length": "436",
            "Accept": "application/json",
            "Connection": "Keep-Alive",
            "User-Agent": "Dalvik/2.1.0 (Linux; U; Android 7.1.2)",
        }
        body = {
            "client_id": "71b963c1b7b6d119",
            "session_token": session_token,
            "grant_type": (
                "urn:ietf:params:oauth:grant-type:jwt-bearer-session-token"
            ),
        }
        uri = "https://accounts.nintendo.com/connect/1.0.0/api/token"
        return self.session.post(uri, headers=header, json=body)

    def get_user_info(self, user_access_token: str) -> dict[str, str]:
        """Gets the user information from the session token.

        Args:
            user_access_token (str): The user access token.

        Returns:
            dict[str, str]: The user information.
        """
        # Get user information
        url = "https://api.accounts.nintendo.com/2.0.0/users/me"
        header = {
            "User-Agent": "NASDKAPI; Android",
            "Content-Type": "application/json",
            "Accept": "application/json",
            "Authorization": f"Bearer {user_access_token}",
            "Host": "api.accounts.nintendo.com",
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
        }
        response = self.session.get(url, headers=header)
        return response.json()

    def get_gtoken(
        self, session_token: str, f_token_url: str | None = None
    ) -> str:
        """Gets the gtoken from the user access token.

        Args:
            f_token_url (str): The url to get the user access token from.
            session_token (str): The session token.

        Raises:
            NintendoException: If the user access token could not be retrieved.

        Returns:
            str: The gtoken.
        """
        f_token_url = f_token_url if f_token_url is not None else IMINK_URL
        # Get user access token
        user_access_response = self.get_user_access_token(session_token)
        user_access_json = user_access_response.json()
        try:
            self._user_access_token = cast(
                str, user_access_json["access_token"]
            )
            self._id_token = user_access_json["id_token"]
        except (KeyError, TypeError, AttributeError):
            raise NintendoException(
                "Failed to get user access token. "
                + f"Response: {user_access_json}"
            )

        user_info = self.get_user_info(self._user_access_token)
        self._user_info = user_info
        splatoon_response = self.f_token_generation_step_1(
            user_access_response, user_info, f_token_url
        )
        id_token = splatoon_response.json()["result"]["webApiServerCredential"][
            "accessToken"
        ]
        gtoken = self.f_token_generation_step_2(id_token, f_token_url)
        self._gtoken = gtoken
        return gtoken

    def get_ftoken(
        self, f_token_url: str, id_token: str, step: int
    ) -> tuple[str, str, str]:
        """Given the f_token_url, id_token, and step, returns the f_token,
        request_id, and timestamp from the response.

        Args:
            f_token_url (str): URL to use for f token generation.
            id_token (str): ID token from user access token response.
            step (int): Step number, 1 or 2.

        Raises:
            FTokenException: If the f token cannot be retrieved.

        Returns:
            tuple:
                str: The f token.
                str: The request ID.
                str: The timestamp.
        """
        header = {
            "User-Agent": f"splatnet3_scraper/{__version__}",
            "Content-Type": "application/json; charset=utf-8",
        }
        body = {
            "token": id_token,
            "hash_method": step,
        }
        bodystr = json.dumps(body)
        response = self.session.post(f_token_url, headers=header, data=bodystr)
        response_json = response.json()
        try:
            f_token = response_json["f"]
            request_id = response_json["request_id"]
            timestamp = response_json["timestamp"]
        except (KeyError, TypeError, AttributeError):
            raise FTokenException(
                "Failed to get f token. " + f"Response: {response_json}"
            )
        return (f_token, request_id, timestamp)

    @retry(times=1, exceptions=(FTokenException, NintendoException, KeyError))
    def f_token_generation_step_1(
        self,
        user_access_response: requests.Response,
        user_info: dict[str, str],
        f_token_url: str,
    ) -> requests.Response:
        """Given the user access token response and user information, returns
        the response from the first step of f token generation. This step
        is retried once if it fails.

        Args:
            user_access_response (requests.Response): The response from the
                user access token request.
            user_info (dict[str, str]): The user information.
            f_token_url (str): The url to get the f token from.

        Returns:
            requests.Response: The response.
        """
        id_token = user_access_response.json()["id_token"]
        f_token, request_id, timestamp = self.get_ftoken(
            f_token_url, id_token, 1
        )

        body = {
            "parameter": {
                "f": f_token,
                "language": user_info["language"],
                "naBirthday": user_info["birthday"],
                "naCountry": user_info["country"],
                "naIdToken": id_token,
                "requestId": request_id,
                "timestamp": timestamp,
            }
        }
        splatoon_response = self.get_splatoon_token(body)
        # Check if the response is valid
        splatoon_response.json()["result"]["webApiServerCredential"][
            "accessToken"
        ]
        return splatoon_response

    @retry(times=1, exceptions=(FTokenException, NintendoException, KeyError))
    def f_token_generation_step_2(
        self,
        id_token: str,
        f_token_url: str,
    ) -> str:
        """Final step of gtoken generation. This step is retried once if it
        fails.

        Args:
            id_token (str): splatoon id token
            f_token_url (str): The url to get the f token from.

        Returns:
            str: The gtoken from the response.
        """
        f_token, request_id, timestamp = self.get_ftoken(
            f_token_url, id_token, 2
        )

        header = {
            "X-Platform": "Android",
            "X-ProductVersion": self.version,
            "Authorization": f"Bearer {id_token}",
            "Content-Type": "application/json; charset=utf-8",
            "Content-Length": "391",
            "Accept-Encoding": "gzip",
            "User-Agent": f"com.nintendo.znca/{self.version}(Android/7.1.2)",
        }
        body = {
            "parameter": {
                "f": f_token,
                "id": 4834290508791808,
                "registrationToken": id_token,
                "requestId": request_id,
                "timestamp": timestamp,
            }
        }
        url = "https://api-lp1.znc.srv.nintendo.net/v2/Game/GetWebServiceToken"
        response = self.session.post(url, headers=header, json=body).json()
        gtoken = cast(str, response["result"]["accessToken"])
        return gtoken

    def get_splatoon_token(self, body: dict) -> requests.Response:
        f_token = body["parameter"]["f"]
        header = {
            "X-Platform": "Android",
            "X-ProductVersion": self.version,
            "Content-Type": "application/json; charset=utf-8",
            "Content-Length": str(990 + len(f_token)),
            "Connection": "Keep-Alive",
            "Accept-Encoding": "gzip",
            "User-Agent": "com.nintendo.znca/"
            + self.version
            + "(Android/7.1.2)",
        }
        url = "https://api-lp1.znc.srv.nintendo.net/v3/Account/Login"
        return self.session.post(url, headers=header, json=body)

    def get_bullet_token(
        self,
        gtoken: str,
        user_info: dict,
        user_agent: str | None = None,
    ) -> str:
        """Given the gtoken and user information, returns the bullet token.

        Bullet token is given by Splatnet and is required to access Splatnet.

        Args:
            gtoken (str): GameWebToken, also known as gtoken. Given by Nintendo.
            user_info (dict): User information. Given by Nintendo.
            user_agent (str | None): User agent. If None, the default user agent
                will be used. Defaults to None.

        Raises:
            SplatnetException: If the provided gtoken is invalid.
            SplatnetException: If the version of Splatnet is obsolete.
            SplatnetException: If no content is returned, indicating that the
                user has not accessed online services before.
            NintendoException: If Nintendo returns an invalid response.

        Returns:
            str: The bullet token.
        """
        user_agent = (
            user_agent if user_agent is not None else DEFAULT_USER_AGENT
        )
        header = {
            "Content-Length": "0",
            "Content-Type": "application/json",
            "Accept-Language": user_info["language"],
            "User-Agent": user_agent,
            "X-Web-View-Ver": self.splatnet_web_version,
            "X-NACOUNTRY": user_info["country"],
            "Accept": "*/*",
            "Origin": SPLATNET_URL,
            "X-Requested-With": "com.nintendo.znca",
        }
        cookies = {
            "_gtoken": gtoken,
            "_dnt": "1",
        }
        url = SPLATNET_URL + "/api/bullet_tokens"
        response = self.session.post(url, headers=header, cookies=cookies)

        if response.status_code == 401:
            raise SplatnetException("Invalid gtoken")
        elif response.status_code == 403:
            raise SplatnetException("Obsolete Version")
        elif response.status_code == 204:
            raise SplatnetException("No Content")

        try:
            return response.json()["bulletToken"]
        except KeyError:
            raise NintendoException("Invalid response from Nintendo")

    @property
    def splatnet_web_version(self) -> str:
        if self._web_view_version is not None:
            return self._web_view_version
        try:
            web_version = get_splatnet_web_version()
            self._web_view_version = web_version
            return web_version
        except SplatnetException as e:
            logger.log(str(e), "warning")
            logger.log("Using fallback web view version", "warning")
            return WEB_VIEW_VERSION_FALLBACK



from __future__ import annotations

import json

import requests

from splatnet3_scraper.constants import (
    DEFAULT_USER_AGENT,
    GRAPH_QL_REFERENCE_URL,
    GRAPHQL_URL,
    SPLATNET_URL,
)
from splatnet3_scraper.utils import get_splatnet_web_version


class GraphQLQueries:
    """Class that contains the GraphQL queries used by splatnet3_scraper.

    Attributes:
        hash_map (dict[str, str]): The hashes for the GraphQL queries.
    """

    def __init__(self) -> None:
        """Initializes the class."""
        self.session = requests.Session()
        self.hash_map = self.get_hashes()

    def get_hashes(self) -> dict[str, str]:
        """Gets the hashes for the GraphQL queries.

        Returns:
            dict[str, str]: The hashes for the GraphQL queries.
        """
        response = self.session.get(GRAPH_QL_REFERENCE_URL)
        hash_map = response.json()["graphql"]["hash_map"]
        return hash_map

    def get_query(self, query_name: str) -> str:
        """Gets a GraphQL query.

        Args:
            query_name (str): The name of the query.

        Returns:
            str: The GraphQL query.
        """
        return self.hash_map[query_name]

    def query_header(
        self,
        bullet_token: str,
        language: str,
        user_agent: str | None = None,
        override: dict[str, str] = {},
    ) -> dict[str, str]:
        """Gets the headers for the GraphQL queries.

        Args:
            bullet_token (str): The bullet token.
            language (str): The language code to use.
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            override (dict[str, str]): The headers to override. Defaults to an
                empty dictionary.

        Returns:
            dict[str, str]: The headers for the GraphQL queries.
        """

        if user_agent is None:
            user_agent = DEFAULT_USER_AGENT

        headers = {
            "Authorization": f"Bearer {bullet_token}",
            "Accept-Language": language,
            "User-Agent": user_agent,
            "X-Web-View-Ver": get_splatnet_web_version(),
            "Content-Type": "application/json",
            "Accept": "*/*",
            "Origin": SPLATNET_URL,
            "X-Requested-With": "com.nintendo.znca",
            "Referer": (
                f"{SPLATNET_URL}?"
                f"lang={language}"
                f"&na_country={language[-2:]}"
                f"&na_lang={language}"
            ),
            "Accept-Encoding": "gzip, deflate",
        }
        headers.update(override)
        return headers

    def query_body_hash(
        self, query_hash: str | bytes, variables: dict[str, str] = {}
    ) -> str:
        """Gets the body for the GraphQL queries, as a string.

        Args:
            query_hash (str | bytes): The hash of the query.
            variables (dict[str, str]): The variables for the query.

        Returns:
            str: The body for the GraphQL queries, as a string.
        """
        out = {
            "extensions": {
                "persistedQuery": {
                    "sha256Hash": query_hash,
                    "version": 1,
                }
            },
            "variables": variables,
        }
        return json.dumps(out)

    def query_body(
        self, query_name: str, variables: dict[str, str] = {}
    ) -> str:
        """Gets the body for the GraphQL queries, as a string.

        Args:
            query_name (str): The name of the query.
            variables (dict[str, str]): The variables for the query.

        Returns:
            str: The body for the GraphQL queries, as a string.
        """
        query = self.get_query(query_name)
        return self.query_body_hash(query, variables)

    def query_hash(
        self,
        query_hash: str | bytes,
        bullet_token: str,
        gtoken: str,
        language: str,
        user_agent: str | None = None,
        variables: dict[str, str] = {},
        override: dict[str, str] = {},
    ) -> requests.Response:
        """Makes a GraphQL query using the persisted query hash.

        Args:
            query_hash (str | bytes): The hash of the query.
            bullet_token (str): The bullet token.
            gtoken (str): The gtoken.
            language (str): The language code to use.
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            variables (dict[str, str]): The variables for the query. Defaults to
                an empty dictionary.
            override (dict[str, str]): The headers to override. Defaults to an
                empty dictionary.

        Returns:
            requests.Response: The response from the query.
        """
        header = self.query_header(bullet_token, language, user_agent, override)
        body = self.query_body_hash(query_hash, variables)
        cookies = {
            "_gtoken": gtoken,
        }
        return self.session.post(
            GRAPHQL_URL, headers=header, data=body, cookies=cookies
        )

    def query(
        self,
        query_name: str,
        bullet_token: str,
        gtoken: str,
        language: str,
        user_agent: str | None = None,
        variables: dict[str, str] = {},
        override: dict[str, str] = {},
    ) -> requests.Response:
        """Makes a GraphQL query using the query name to get the hash.

        Args:
            query_name (str): The name of the query.
            bullet_token (str): The bullet token.
            gtoken (str): The gtoken.
            language (str): The language code to use.
            user_agent (str | None): The user agent to use. If None, the default
                user agent will be used. Defaults to None.
            variables (dict[str, str]): The variables for the query. Defaults to
                an empty dictionary.
            override (dict[str, str]): The headers to override. Defaults to an
                empty dictionary.

        Returns:
            requests.Response: The response from the query.
        """
        query = self.get_query(query_name)
        return self.query_hash(
            query,
            bullet_token,
            gtoken,
            language,
            user_agent,
            variables,
            override,
        )


queries = GraphQLQueries()



from typing import Any, Literal, overload

from splatnet3_scraper.scraper.json_parser import JSONParser


class QueryResponse:
    """Represents a response from the API. This is a thin wrapper around the
    JSONParser class.
    """

    def __init__(
        self,
        data: dict[str, Any] | list[dict[str, Any]],
        additional_data: list[dict[str, Any]] | dict[str, Any] | None = None,
    ) -> None:
        """Initializes a QueryResponse.

        Args:
            data (dict[str, Any]): The data from the response.
            additional_data (list[dict[str, Any]] | None): Any additional data,
                useful for complex queries that require multiple requests.
                Defaults to None.
        """
        self._data = data
        self._additional_data = additional_data

    @property
    def data(self) -> dict[str, Any] | list[dict[str, Any]]:
        """The data from the response.

        Returns:
            dict[str, Any]: The data.
        """
        return self._data

    @property
    def additional_data(self) -> list[dict[str, Any]] | dict[str, Any]:
        """Any additional data, useful for complex queries that require
            multiple requests.

        Raises:
            AttributeError: If there is no additional data.

        Returns:
            list[dict[str, Any]] | dict[str, Any]: The additional data.
        """
        if self._additional_data is None:
            raise AttributeError("No additional data")
        return self._additional_data

    def parse_json(self, additional: bool = False) -> JSONParser:
        """The JSONParser object containing the data.

        Args:
            additional (bool): Whether to parse the additional data. Defaults
                to False.

        Raises:
            AttributeError: If there is no additional data and additional is
                True.

        Returns:
            JSONParser: The data.
        """
        if additional and self._additional_data is None:
            raise AttributeError("No additional data")
        elif additional:
            return JSONParser(self._additional_data)
        else:
            return JSONParser(self._data)

    def __repr__(self) -> str:
        detail_str = "+" if self._additional_data is not None else ""
        return f"QueryResponse{detail_str}()"

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, QueryResponse):
            return False
        return (
            self._data == other._data
            and self._additional_data == other._additional_data
        )

    def __getitem__(
        self, key: str | int | tuple[str | int, ...]
    ) -> "QueryResponse":
        if isinstance(key, tuple):
            for k in key:
                self = self[k]
            return self

        data = self._data[key]  # type: ignore
        if isinstance(data, (dict, list)):
            return QueryResponse(data)
        return data

    def keys(self) -> list[str]:
        """Returns a list of keys in the data.

        Returns:
            list[str]: The keys in the data.
        """
        return list(self._data.keys())

    def values(self) -> list[Any]:
        """Returns a list of values in the data.

        Returns:
            list[Any]: The values in the data.
        """
        return list(self._data.values())

    def items(self) -> list[tuple[str, Any]]:
        """Returns a list of items in the data.

        Returns:
            list[tuple[str, Any]]: The items in the data.
        """
        return list(self._data.items())

    def __iter__(self) -> iter:
        return iter(self._data)

    @overload
    def show(self, return_value: bool = Literal[False]) -> None:
        ...

    @overload
    def show(self, return_value: bool = Literal[True]) -> dict[str, Any]:
        ...

    @overload
    def show(self, return_value: bool = ...) -> None | dict[str, Any]:
        ...

    def show(self, return_value: bool = False) -> None | dict[str, Any]:
        """Prints the data to the console. If return_value is True, returns the
        data as a dict instead.

        Args:
            return_value (bool): Whether to return the data as a dict. If True,
                the data will be returned instead of printed. Defaults to False.

        Returns:
            None | dict[str, Any]: The data as a dict if return_value is True,
                otherwise None.
        """
        if return_value:
            return self._data
        print(self._data)



class NintendoException(Exception):
    """Base class for all Nintendo exceptions."""

    pass


class FTokenException(Exception):
    """Base class for all Imink exceptions."""

    pass


class SplatnetException(Exception):
    """Base class for all Splatnet exceptions."""

    pass



import re
from functools import cache
from typing import Any, Callable, ParamSpec, Type, TypeVar

import requests

from splatnet3_scraper.constants import GRAPH_QL_REFERENCE_URL
from splatnet3_scraper.logs import logger

T = TypeVar("T")
P = ParamSpec("P")

json_splitter_re = re.compile(r"[\;\.]")


def retry(
    times: int,
    exceptions: tuple[Type[Exception], ...] | Type[Exception] = Exception,
    call_on_fail: Callable[[], None] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """Decorator that retries a function a specified number of times if it
    raises a specific exception or tuple of exceptions.

    Args:
        times (int): Max number of times to retry the function before raising
            the exception.
        exceptions (tuple[Type[Exception], ...] | Type[Exception]): Exception
            or tuple of exceptions to catch. Defaults to Exception.
        call_on_fail (Callable[[], None] | None): Function to call if the
            function fails. If None, nothing will be called. Defaults to None.

    Returns:
        Callable[[Callable[P, T]], Callable[P, T]]: Decorator.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            for i in range(times):
                try:
                    return func(*args, **kwargs)
                except exceptions:
                    logger.log(
                        f"{func.__name__} failed on attempt {i + 1} of "
                        f"{times + 1}, retrying."
                    )
                    if call_on_fail is not None:
                        logger.log(f"Calling {call_on_fail.__name__}...")
                        call_on_fail()

            return func(*args, **kwargs)

        return wrapper

    return decorator


@cache
def get_splatnet_web_version() -> str:
    """Gets the web view version from the GraphQL reference.

    Returns:
        str: The web view version.
    """
    response = requests.get(GRAPH_QL_REFERENCE_URL)
    return response.json()["version"]


def linearize_json(
    json_data: dict[str, Any]
) -> tuple[tuple[str, ...], list[Any]]:
    """Linearizes a JSON object.

    Args:
        json_data (dict[str, Any]): The JSON object to linearize.

    Returns:
        tuple:
            tuple[str, ...]: The keys of the JSON object.
            list[Any]: The values of the JSON object.
    """
    keys = []
    values = []

    for key, value in json_data.items():
        if isinstance(value, dict):
            sub_keys, sub_values = linearize_json(value)
            keys.extend([(key + "." + sub_key) for sub_key in sub_keys])
            values.extend(sub_values)
        elif isinstance(value, list):
            for i, item in enumerate(value):
                if isinstance(item, dict):
                    sub_keys, sub_values = linearize_json(item)
                    keys.extend(
                        [
                            (key + ";" + str(i) + "." + sub_key)
                            for sub_key in sub_keys
                        ]
                    )
                    values.extend(sub_values)
                else:
                    keys.append(key + ";" + str(i))
                    values.append(item)
        else:
            keys.append(key)
            values.append(value)

    # Turn the keys into an immutable tuple so it can be hashed
    out_keys = tuple(keys)
    return out_keys, values


def delinearize_json(
    keys: list[str] | tuple[str, ...], values: list[Any]
) -> dict[str, Any]:
    """Delinearizes a JSON object.

    Args:
        keys (list[str]): The keys of the JSON object.
        values (list[Any]): The values of the JSON object.

    Returns:
        dict[str, Any]: The JSON object.
    """
    json_data = {}

    # Sort the keys by depth
    depths = [len(json_splitter_re.split(key)) for key in keys]
    kvd = list(zip(keys, values, depths))
    kvd = sorted(kvd, key=lambda x: (x[2], x[0]))
    keys, values, _ = list(zip(*kvd))

    # Delinearize
    for key, value in zip(keys, values):
        # If the key is split by a period, it's a nested object. If it's split
        # by a semicolon, it's a list. Check which one is first.
        subkeys = json_splitter_re.split(key)
        splitters = json_splitter_re.findall(key)
        if len(subkeys) == 1:
            json_data[key] = value
            continue
        # If the key is split by a semicolon, turn the next key value into an
        # integer
        for i, splitter in enumerate(splitters):
            if splitter == ";":
                subkeys[i + 1] = int(subkeys[i + 1])

        current = json_data
        for i, splitter in enumerate(splitters):
            # If the key already exists, move on to the next key
            if isinstance(current, (list, dict)):
                chosen_subkey = subkeys[i]
                if isinstance(current, list):
                    condition = len(current) > chosen_subkey
                elif isinstance(current, dict):
                    condition = chosen_subkey in current

                if condition:
                    next_obj = current[chosen_subkey]
                    # Next object might be None as an artifact of header merging
                    if next_obj is not None:
                        current = next_obj
                        continue
            new_obj: dict | list = {} if (splitter == ".") else []

            # Append or assign the new object to the current object depending
            # on type.
            if isinstance(current, list):
                current.append(new_obj)
            elif isinstance(current, dict):
                current[subkeys[i]] = new_obj
            # Mypy doesn't like that this is dynamically typed, so we have to
            # ignore it. This is python, not C.
            current = new_obj  # type: ignore
        if isinstance(current, list):
            current.append(value)
        else:
            current[subkeys[-1]] = value

    return json_data



class QueryMap:
    BANKARA_BATTLE_HISTORIES = "BankaraBattleHistoriesQuery"
    BANKARA_BATTLE_HISTORIES_REFETCH = "BankaraBattleHistoriesRefetchQuery"
    BATTLE_HISTORY_CURRENT_PLAYER = "BattleHistoryCurrentPlayerQuery"
    CATALOG = "CatalogQuery"
    CHECKIN = "CheckinQuery"
    CHECKIN_QR = "CheckinWithQRCodeMutation"
    CONFIGURE_ANALYTICS = "ConfigureAnalyticsQuery"
    COOP_HISTORY_DETAIL = "CoopHistoryDetailQuery"
    COOP_HISTORY_DETAIL_REFETCH = "CoopHistoryDetailRefetchQuery"
    COOP_HISTORY = "CoopHistoryQuery"
    COOP_PAGER_LATEST_COOP = "CoopPagerLatestCoopQuery"
    COOP_RECORD_BIG_RUN_RECORD_CONTAINER_PAGINATION = (
        "CoopRecordBigRunRecordContainerPaginationQuery"
    )
    COOP_RECORD = "CoopRecordQuery"
    COOP_RECORD_REFETCH = "CoopRecordRefetchQuery"
    CREATE_MY_OUTFIT_MUTATION = "CreateMyOutfitMutation"
    DETAIL_FEST_RECORD_DETAIL = "DetailFestRecordDetailQuery"
    DETAIL_FEST_REFETCH = "DetailFestRefethQuery"
    DETAIL_FEST_VOTING_STATUS_REFETCH = "DetailFestVotingStatusRefethQuery"
    DETAIL_RANKING = "DetailRankingQuery"
    DETAIL_TAB_VIEW_WEAPON_TOPS_AR_REFETCH = (
        "DetailTabViewWeaponTopsArRefetchQuery"
    )
    DETAIL_TAB_VIEW_WEAPON_TOPS_CL_REFETCH = (
        "DetailTabViewWeaponTopsClRefetchQuery"
    )
    DETAIL_TAB_VIEW_WEAPON_TOPS_GL_REFETCH = (
        "DetailTabViewWeaponTopsGlRefetchQuery"
    )
    DETAIL_TAB_VIEW_WEAPON_TOPS_LF_REFETCH = (
        "DetailTabViewWeaponTopsLFRefetchQuery"
    )
    DETAIL_TAB_VIEW_X_RANKING_AR_REFETCH = "DetailTabViewXRankingArRefetchQuery"
    DETAIL_TAB_VIEW_X_RANKING_CL_REFETCH = "DetailTabViewXRankingClRefetchQuery"
    DETAIL_TAB_VIEW_X_RANKING_GL_REFETCH = "DetailTabViewXRankingGlRefetchQuery"
    DETAIL_TAB_VIEW_X_RANKING_LF_REFETCH = "DetailTabViewXRankingLFRefetchQuery"
    DETAIL_VOTING_STATUS = "DetailVotingStatusQuery"
    DOWNLOAD_SEARCH_REPLAY = "DownloadSearchReplayQuery"
    FEST_RECORD = "FestRecordQuery"
    FEST_RECORD_REFETCH = "FestRecordRefetchQuery"
    FRIEND_LIST = "FriendListQuery"
    FRIEND_LIST_REFETCH = "FriendListRefetchQuery"
    GESOTOWN = "GesotownQuery"
    GESOTOWN_REFETCH = "GesotownRefetchQuery"
    HERO_HISTORY = "HeroHistoryQuery"
    HERO_HISTORY_REFETCH = "HeroHistoryRefetchQuery"
    HISTORY_RECORD = "HistoryRecordQuery"
    HISTORY_RECORD_REFETCH = "HistoryRecordRefetchQuery"
    HOME = "HomeQuery"
    JOURNEY_CHALLENGE_DETAIL = "JourneyChallengeDetailQuery"
    JOURNEY_CHALLENGE_DETAIL_REFETCH = "JourneyChallengeDetailRefetchQuery"
    JOURNEY = "JourneyQuery"
    JOURNEY_REFETCH = "JourneyRefetchQuery"
    LATEST_BATTLE_HISTORIES = "LatestBattleHistoriesQuery"
    LATEST_BATTLE_HISTORIES_REFETCH = "LatestBattleHistoriesRefetchQuery"
    MY_OUTFITS_DETAIL = "MyOutfitsDetailQuery"
    MY_OUTFITS = "MyOutfitsQuery"
    MY_OUTFITS_REFETCH = "MyOutfitsRefetchQuery"
    PAGER_LATEST_VS_DETAIL = "PagerLatestVsDetailQuery"
    PAGER_UPDATE_BATTLE_HISTORIES_BY_VS_MODE = (
        "PagerUpdateBattleHistoriesByVsModeQuery"
    )
    PHOTO_ALBUM = "PhotoAlbumQuery"
    PHOTO_ALBUM_REFETCH = "PhotoAlbumRefetchQuery"
    PRIVATE_BATTLE_HISTORIES = "PrivateBattleHistoriesQuery"
    PRIVATE_BATTLE_HISTORIES_REFETCH = "PrivateBattleHistoriesRefetchQuery"
    RANKING_HOLDERS_FEST_TEAM_RANKING_HOLDERS_PAGINATION = (
        "RankingHoldersFestTeamRankingHoldersPaginationQuery"
    )
    REGULAR_BATTLE_HISTORIES = "RegularBattleHistoriesQuery"
    REGULAR_BATTLE_HISTORIES_REFETCH = "RegularBattleHistoriesRefetchQuery"
    REPLAY_MODAL_RESERVE_REPLAY_DOWNLOAD_MUTATION = (
        "ReplayModalReserveReplayDownloadMutation"
    )
    REPLAY = "ReplayQuery"
    REPLAY_UPLOADED_REPLAY_LIST_REFETCH = "ReplayUploadedReplayListRefetchQuery"
    SALE_GEAR_DETAIL_ORDER_GESOTOWN_GEAR_MUTATION = (
        "SaleGearDetailOrderGesotownGearMutation"
    )
    SALE_GEAR_DETAIL = "SaleGearDetailQuery"
    SETTING = "SettingQuery"
    STAGE_RECORD = "StageRecordQuery"
    STAGE_RECORDS_REFETCH = "StageRecordsRefetchQuery"
    STAGE_SCHEDULE = "StageScheduleQuery"
    SUPPORT_BUTTON_SUPPORT_CHALLENGE_MUTATION = (
        "SupportButton_SupportChallengeMutation"
    )
    UPDATE_MY_OUTFIT_MUTATION = "UpdateMyOutfitMutation"
    VOTES_UPDATE_FEST_VOTE_MUTATION = "VotesUpdateFestVoteMutation"
    VS_HISTORY_DETAIL_PAGER_REFETCH = "VsHistoryDetailPagerRefetchQuery"
    VS_HISTORY_DETAIL = "VsHistoryDetailQuery"
    WEAPON_RECORD = "WeaponRecordQuery"
    WEAPON_RECORDS_REFETCH = "WeaponRecordsRefetchQuery"
    X_BATTLE_HISTORIES = "XBattleHistoriesQuery"
    X_BATTLE_HISTORIES_REFETCH = "XBattleHistoriesRefetchQuery"
    X_RANKING_DETAIL = "XRankingDetailQuery"
    X_RANKING_DETAIL_REFETCH = "XRankingDetailRefetchQuery"
    X_RANKING = "XRankingQuery"
    X_RANKING_REFETCH = "XRankingRefetchQuery"
    MY_OUTFIT_COMMON_DATA_EQUIPMENTS = "myOutfitCommonDataEquipmentsQuery"
    MY_OUTFIT_COMMON_DATA_FILTERING_CONDITION = (
        "myOutfitCommonDataFilteringConditionQuery"
    )
    REFETCHABLE_COOP_HISTORY_COOP_RESULT = (
        "refetchableCoopHistory_coopResultQuery"
    )
    USE_CURRENT_FEST = "useCurrentFestQuery"

    # Convenience aliases
    VS_DETAIL = "VsHistoryDetailQuery"
    SALMON_DETAIL = "CoopHistoryDetailQuery"
    ANARCHY = "BankaraBattleHistoriesQuery"
    REGULAR = "RegularBattleHistoriesQuery"
    XBATTLE = "XBattleHistoriesQuery"
    PRIVATE = "PrivateBattleHistoriesQuery"
    LATEST = "LatestBattleHistoriesQuery"
    SALMON = "CoopHistoryQuery"
    TURF = "RegularBattleHistoriesQuery"
    COOP = "CoopHistoryQuery"
    ANARCHY_DETAIL = "VsHistoryDetailQuery"
    REGULAR_DETAIL = "VsHistoryDetailQuery"
    X_DETAIL = "VsHistoryDetailQuery"
    PRIVATE_DETAIL = "VsHistoryDetailQuery"
    LATEST_DETAIL = "VsHistoryDetailQuery"
    COOP_DETAIL = "CoopHistoryDetailQuery"

    @staticmethod
    def get(query: str) -> str:
        """Gets the query from the query map.

        Args:
            query (str): The query to get.

        Returns:
            str: The query.
        """
        return getattr(QueryMap, query.upper())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splatnet3_scraper',
 'splatnet3_scraper.base',
 'splatnet3_scraper.base.tokens',
 'splatnet3_scraper.scraper']

package_data = \
{'': ['*']}

install_requires = \
['requests>=2.28.1,<3.0.0']

extras_require = \
{'examples': ['pandas[examples]>=1.5.3,<2.0.0',
              'sqlalchemy[examples]>=2.0.1,<3.0.0',
              'psycopg2[examples]>=2.9.5,<3.0.0'],
 'parquet': ['pyarrow[parquet]>=10.0.1,<11.0.0']}

setup_kwargs = {
    'name': 'splatnet3-scraper',
    'version': '0.4.5',
    'description': 'Scraper for SplatNet 3 for Splatoon 3',
    'long_description': '# SplatNet 3 Scraper\n\n[![Tests Status](./reports/junit/tests-badge.svg?dummy=8484744)](https://htmlpreview.github.io/?https://github.com/cesaregarza/SplatNet3_Scraper/blob/main/reports/junit/report.html) ![Coverage Status](./reports/coverage/coverage-badge.svg?dummy=8484744) ![Flake8 Status](./reports/flake8/flake8-badge.svg?dummy=8484744) [![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\n**SplatNet 3 Scraper** is a Python library for scraping data from the Splatoon 3 SplatNet 3 API. It is designed to be as lightweight as possible, with minimal dependencies to make it easy to integrate into other projects.\n\n**SplatNet 3 Scraper** started as a fork of **[s3s](https://github.com/frozenpandaman/s3s)**, but has since been rewritten from scratch while incorporating much of the login flow logic of s3s. As a result, I am deeply indebted to the authors of s3s for their work. This project would not have been possible without their efforts.\n\n## Features\n\n* Lightweight and minimal dependencies. Only requires the `requests` library. Requires Python 3.10 or later.\n* The `scraper` module provides a high level API that enables a quick and easy way to get data from the SplatNet 3 API, only requiring the user to provide their session token.\n* The `base` module provides a low level API that allows for more fine-grained control over the scraping process. It is designed to be used by the scraper module, but is designed to be flexible enough to be used by other projects as well.\n* Configuration file support is compatible with the configuration file format used by `s3s`.\n* Responses from the SplatNet 3 API can be saved and loaded from disk, currently supporting the following formats:\n  * JSON\n  * gzip-compressed JSON\n  * csv\n  * parquet (by installing `splatnet3_scraper[parquet]` or the `pyarrow` library)\n\n## Installation\n\n**SplatNet3_Scraper** is currently under active development but is currently available on PyPI. It can be installed using pip:\n\n```bash\npip install splatnet3_scraper\n```\n\nNote that the current versions of **SplatNet3_Scraper** are currently `v0.x.y`, which means that the API is not guaranteed to be stable and may change in the future with the release of `v1.0.0`.\n\n## Usage\n\nThere are two ways to use **SplatNet3_Scraper**. The first is to use the `scraper` module, which provides a high level API that greatly simplifies the process of retrieving data from SplatNet 3. The second is to use the `base` module, which provides a low level API that allows for much more fine-grained control over the scraping process. Either way, both modules require a session token to be provided.\n\n### Using the `scraper` module\n\nThe `scraper` module is a batteries-included module that allows queries to be made to the SplatNet 3 API with minimal effort. It is designed to be used by the end user, and as such it is the easiest and recommended way to get started with **SplatNet3_Scraper**. The `scraper` module provides the `SplatNet3_Scraper` class, which is used to make queries to the SplatNet 3 API. The `SplatNet3_Scraper` class can be instantiated in one of a few ways: by providing a session token, by providing the path to a configuration file, or by loading environment variables.\n\n#### Instantiating the `SplatNet3_Scraper` class by providing a session token\n\n```python\nfrom splatnet3_scraper import SplatNet3_Scraper\nscraper = SplatNet3_Scraper.from_session_token("session_token")\nscraper.query("StageScheduleQuery")\n```\n\n#### Instantiating the `SplatNet3_Scraper` class by providing the path to a configuration file\n\n```python\nfrom splatnet3_scraper import SplatNet3_Scraper\nscraper = SplatNet3_Scraper.from_config_file(".splatnet3_scraper")\nscraper.query("StageScheduleQuery")\n```\n\n#### Instantiating the `SplatNet3_Scraper` class by loading environment variables\n\nThe following environment variables are supported:\n\n* `SN3S_SESSION_TOKEN`\n* `SN3S_GTOKEN`\n* `SN3S_BULLET_TOKEN`\n\n```python\nfrom splatnet3_scraper import SplatNet3_Scraper\nscraper = SplatNet3_Scraper.from_env()\nscraper.query("StageScheduleQuery")\n```\n\n#### Querying the SplatNet 3 API\n\nThe `SplatNet3_Scraper` class provides a `query` method that can be used to make queries to the SplatNet 3 API. The `query` method takes a single argument, which is the name of the query to make. The `query` method returns a `QueryResponse` object, which contains the response data from the SplatNet 3 API. The `QueryResponse` object provides a `data` property that can be used to access the response data. The `QueryResponse` module also supports numpy-style indexing, which can be used to quickly and clearly access specific parts of the response data. For example, the following code will print the game mode name of the the current stage rotation schedule:\n\n```python\nfrom splatnet3_scraper import SplatNet3_Scraper\nscraper = SplatNet3_Scraper.from_env()\nresponse = scraper.query("StageScheduleQuery")\nprint(response["xSchedules", "nodes", 0, "vsRule", "name"])\n```\n\n#### Saving and loading responses\n\nThe `QueryResponse` class provides a `parsed_json` method that can be used to generate a `JSONParser` object from the response data. The `JSONParser` class provides multiple ways of interacting with the given data, including the ability to save the data to disk in a variety of formats. There are currently four different formats that are supported and can be used by passing the desired format to a `to_*` method such as `to_json`. The following formats are supported:\n\n* JSON\n* gzip-compressed JSON\n* csv\n* parquet (by installing `splatnet3_scraper[parquet]` or the `pyarrow` library)\n\nNote: csv and parquet formats work by converting the response data from a nested dictionary to a columnar format. This is not recommended for single queries, but can be useful for interacting with large amounts of data as it deduplicates the JSON structure and allows for more efficient storage and querying.\n\nThe following code will save the response data to a file named `response.json` in the current directory:\n\n```python\nfrom splatnet3_scraper import SplatNet3_Scraper\nscraper = SplatNet3_Scraper.from_env()\nresponse = scraper.query("StageScheduleQuery")\nresponse.parsed_json().to_json("response.json")\n```\n\nAdditionally, the `JSONParser` class provides a `from_*` method that can be used to load data from a file. The following code will load the response data from the file `response.json` in the current directory:\n\n```python\nfrom splatnet3_scraper import JSONParser\nparser = JSONParser.from_json("response.json")\n```\n\n## Symbols\n\n| Symbol | Meaning |\n| ------ | ------- |\n| :white_check_mark: | Implemented |\n| :construction: | In progress |\n| :world_map: | Planned |\n| :x: | Not planned |\n\n## Roadmap\n\n| Feature | Status |\n| ------- | ------ |\n| Support for the SplatNet 3 API | :white_check_mark: |\n| Full support for the SplatNet 3 API | :world_map: |\n| Support for the SplatNet 2 API | :x: |\n| Obtaining session tokens | :white_check_mark: |\n| Full documentation | :world_map: |\n| Full unit test coverage | :white_check_mark: |\n| Columnar data format support | :construction: |\n| CLI interface | :x: |\n| Integration with stat.ink | :x: |\n| PyPI package | :white_check_mark: |\n| Docker image | :world_map: |\n| Executable binary | :x: |\n\n## Docker Note\n\nThis project currently uses the standard library heavily, and as such it is not compatible with the `python:alpine` Docker image. I have no plans to change this. Use the `python:slim` image instead.\n\nSplatNet3_Scraper is licensed under the GPLv3. See the LICENSE file for more details.\n',
    'author': 'Cesar E Garza',
    'author_email': 'cesar@cegarza.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



# Not currently implemented, placeholder to reduce refactoring later on


class Logger:
    def __init__(self, log_file_name: str) -> None:
        self.log_file_name = log_file_name

    def log(self, message: str, *args) -> None:
        print(message)


logger = Logger("splatnet3_scraper.log")



SPLATNET_URL = "https://api.lp1.av5ja.srv.nintendo.net"
GRAPHQL_URL = SPLATNET_URL + "/api/graphql"
GRAPH_QL_REFERENCE_URL = (
    "https://raw.githubusercontent.com"
    "/imink-app/SplatNet3/master/Data/splatnet3_webview_data.json"
)
IOS_APP_URL = (
    "https://apps.apple.com/us/app/nintendo-switch-online/id1234806557"
)
IMINK_URL = "https://api.imink.app/f"
DEFAULT_USER_AGENT = (
    "Mozilla/5.0 (Linux; Android 11; Pixel 5) "
    + "AppleWebKit/537.36 (KHTML, like Gecko) "
    + "Chrome/94.0.4606.61 "
    + "Mobile Safari/537.36"
)
WEB_VIEW_VERSION_FALLBACK = "2.0.0-bd36a652"
APP_VERSION_FALLBACK = "2.4.0"


class TOKENS:
    SESSION_TOKEN = "session_token"
    GTOKEN = "gtoken"
    BULLET_TOKEN = "bullet_token"


TOKEN_EXPIRATIONS = {
    TOKENS.GTOKEN: (60 * 60 * 6) + (60 * 30),
    TOKENS.BULLET_TOKEN: (60 * 60 * 2),
}
ENV_VAR_NAMES = {
    TOKENS.SESSION_TOKEN: "SN3S_SESSION_TOKEN",
    TOKENS.GTOKEN: "SN3S_GTOKEN",
    TOKENS.BULLET_TOKEN: "SN3S_BULLET_TOKEN",
}

PRIMARY_ONLY = [
    "Comeback",
    "Last-Ditch Effort",
    "Opening Gambit",
    "Tenacity",
    "Ability Doubler",
    "Haunt",
    "Ninja Squid",
    "Respawn Punisher",
    "Thermal Ink",
    "Drop Roller",
    "Object Shredder",
    "Stealth Jump",
]

ABILITIES = [
    "Ink Saver (Main)",
    "Ink Saver (Sub)",
    "Ink Recovery Up",
    "Run Speed Up",
    "Swim Speed Up",
    "Special Charge Up",
    "Special Saver",
    "Special Power Up",
    "Quick Respawn",
    "Quick Super Jump",
    "Sub Power Up",
    "Ink Resistance Up",
    "Sub Resistance Up",
    "Intensify Action",
]

ALL_ABILITIES = PRIMARY_ONLY + ABILITIES
